from aws_syncr.amazon.common import AmazonMixin
from aws_syncr.differ import Differ

from contextlib import contextmanager
import logging
import base64
import json
import uuid
import six

log = logging.getLogger("aws_syncr.amazon.lambdas")

class Lambdas(AmazonMixin, object):
    def __init__(self, amazon, environment, accounts, dry_run):
        self.amazon = amazon
        self.dry_run = dry_run

        self.accounts = accounts
        self.account_id = accounts[environment]
        self.environment = environment

    def function_info(self, function_name, location):
        with self.ignore_missing():
            return self.amazon.session.client('lambda', location).get_function(FunctionName=function_name)

    @contextmanager
    def code_options(self, code):
        options = {}
        if code.s3_address:
            options["Key"] = code.key
            options["S3Bucket"] = code.bucket
            if code.version is not NotSpecified:
                options["S3ObjectVersion"] = code.version
            yield options
        else:
            with code.zipfile() as location:
                if location:
                    yield {"ZipFile": open(location, 'rb').read()}

    def create_function(self, name, description, location, runtime, role, handler, timeout, memory_size, code):
        client = self.amazon.session.client('lambda', location)
        with self.catch_boto_400("Couldn't Make function", function=name):
            for _ in self.change("+", "function", function=name):
                kwargs = dict(
                      FunctionName=name, Runtime=runtime, Role=role, Handler=handler
                    , Description = description, Timeout=timeout, MemorySize=memory_size
                    )

                with self.code_options(code) as options:
                    kwargs["Code"] = options
                    client.create_function(**kwargs)

    def modify_function(self, function_info, name, description, location, runtime, role, handler, timeout, memory_size, code):
        client = self.amazon.session.client('lambda', location)

        wanted = dict(
              FunctionName=name, Role=role, Handler=handler
            , Description=description, Timeout=timeout, MemorySize=memory_size
            )

        current = dict((key, function_info["Configuration"][key]) for key in (
              "FunctionName", "Role", "Handler", "Description", "Timeout", "MemorySize"
            )
        )

        changes = list(Differ.compare_two_documents(current, wanted))
        if changes:
            with self.catch_boto_400("Couldn't modify function", function=name):
                for _ in self.change("M", "function", changes=changes, function=name):
                    client.update_function_configuration(**wanted)

    def deploy_function(self, name, code, location):
        client = self.amazon.session.client('lambda', location)
        with self.code_options(code) as options:
            for _ in self.change("D", "function", function=name):
                with self.catch_boto_400("Couldn't deploy function", function=name):
                    return client.update_function_code(FunctionName=name, **options)

    def test_function(self, name, event, location):
        client = self.amazon.session.client('lambda', location)
        log.info("Invoking function %s", name)
        if not isinstance(event, six.string_types):
            event = json.dumps(event)
        res = client.invoke(FunctionName=name, InvocationType="RequestResponse", Payload=event, LogType="Tail")
        res['Payload'] = json.loads(res['Payload'].read().decode('utf-8'))
        if 'LogResult' in res:
            print(base64.b64decode(res['LogResult']).decode('utf-8'))
            del res['LogResult']
        return res

    def modify_resource_policy_for_gateway(self, function_arn, function_location, gateway_arn, gateway_name):
        lambda_client = self.amazon.session.client("lambda", function_location)
        policy = {}
        with self.ignore_missing():
            policy = lambda_client.get_policy(FunctionName = function_arn)["Policy"]
            policy = json.loads(policy)
        statements = policy.get("Statement", [])

        current_apigateway_statements = []

        wanted = {'Resource': function_arn, 'Effect': 'Allow',  'Action': 'lambda:InvokeFunction',  'Principal': {'Service': 'apigateway.amazonaws.com'}}
        for statement in statements:
            if all(wanted[key] == statement[key] for key in wanted):
                current_apigateway_statements.append(statement)

        for current_apigateway_statement in current_apigateway_statements:
            if current_apigateway_statement.get("Condition", {}).get("ArnLike", {}).get("AWS:SourceArn") == gateway_arn:
                # Our work here is done, no changes to make
                return

        # At this point, we have no statement allowing our gateway to invoke our lambda function
        # So let's add a new statement!!!
        new_statement = wanted
        new_statement["Condition"] = {"ArnLike": { 'AWS:SourceArn' : gateway_arn } }

        # Make a copy of the statements with our new statement
        new_statements = list(statements)
        new_statements.append(new_statement)

        # Show the differences to the user
        changes = list(Differ.compare_two_documents(statements, new_statements))
        if changes:
            function_name = function_arn.split(":")[-1]
            for _ in self.change("M", "Lambda resource policy", gateway=gateway_name, function=function_name, changes=changes):
                lambda_client.add_permission(
                      FunctionName=function_name
                    , StatementId = str(uuid.uuid1())
                    , Action = new_statement["Action"]
                    , Principal = new_statement["Principal"]["Service"]
                    , SourceArn = gateway_arn
                    )




from aws_syncr.option_spec.statements import grant_statement_spec, resource_policy_statement_spec
from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.option_spec.documents import Document
from aws_syncr.errors import BadTemplate

from input_algorithms import spec_base as sb
from input_algorithms.spec_base import Spec
from input_algorithms.dictobj import dictobj

from option_merge import MergedOptions
import six

class encryption_keys_spec(Spec):
    def normalise(self, meta, val):
        if 'use' in val:
            template = val['use']
            if template not in meta.everything['templates']:
                available = list(meta.everything['templates'].keys())
                raise BadTemplate("Template doesn't exist!", wanted=template, available=available, meta=meta)

            val = MergedOptions.using(meta.everything['templates'][template], val)

        formatted_string = sb.formatted(sb.string_or_int_as_string_spec(), MergedOptionStringFormatter, expected_type=six.string_types)
        key_name = meta.key_names()['_key_name_0']

        key = sb.create_spec(EncryptionKey
            , name = sb.overridden(key_name)
            , location = sb.required(formatted_string)
            , description = formatted_string
            , grant = sb.listof(grant_statement_spec('key', key_name))
            , admin_users = sb.listof(sb.any_spec())
            ).normalise(meta, val)

        statements = [{"principal": {"iam": "root"}, "action": "kms:*", "resource": "*", "Sid": ""}]
        if key.admin_users:
            for admin_user in key.admin_users:
                statements.append({"principal": admin_user, "action": "kms:*", "resource": { "kms": "__self__" }, "Sid": ""})

        key.policy = sb.container_spec(Document, sb.listof(resource_policy_statement_spec('key', key_name))).normalise(meta.at("admin_users"), statements)
        return key

class EncryptionKeys(dictobj):
    fields = ["items"]

    def sync_one(self, aws_syncr, amazon, key):
        """Make sure this key is as defined"""
        key_info = amazon.kms.key_info(key.name, key.location)
        if not key_info:
            amazon.kms.create_key(key.name, key.description, key.location, key.grant, key.policy.document)
        else:
            amazon.kms.modify_key(key_info, key.name, key.description, key.location, key.grant, key.policy.document)

class EncryptionKey(dictobj):
    fields = {
          'name': "Name of the key"
        , 'location': "The region the key exists in"
        , 'description': "Description of the key"
        , 'admin_users': "The admin_users for this key"
        , 'grant': "The grants given to the key"
        }

def __register__():
    return {(10, "encryption_keys"): sb.container_spec(EncryptionKeys, sb.dictof(sb.string_spec(), encryption_keys_spec()))}




from aws_syncr.option_spec.resources import resource_spec, iam_specs
from aws_syncr.errors import BadOption, BadPolicy

from input_algorithms.spec_base import NotSpecified, apply_validators
from input_algorithms import spec_base as sb, validators
from input_algorithms.dictobj import dictobj
from option_merge import MergedOptions
from itertools import chain
import six

def capitalize(arg):
    if type(arg) is tuple:
        capitalized = ''.join(part.capitalize() for part in arg)
        arg = ''.join(arg)
    else:
        capitalized = arg.capitalize()
    return arg, capitalized

class statement_spec(sb.Spec):
    args = None
    final_kls = None
    required = []
    validators = []
    invalid_args = []

    def setup(self, self_type, self_name):
        self.self_type = self_type
        self.self_name = self_name
        if self.args is None or not self.final_kls:
            raise NotImplementedError("Need to use a subclass of statement_spec that defines args and final_kls")

    def normalise(self, meta, val):
        self.complain_about_invalid_args(meta, val)

        apply_validators(meta, val, self.validators, chain_value=False)
        args, spec = self.make_spec()
        normalised = spec.normalise(meta, val)

        kwargs = self.make_kwargs(meta, args, normalised)
        self.complain_about_missing_args(meta, kwargs)

        return self.final_kls(**kwargs)

    def complain_about_invalid_args(self, meta, val):
        for arg in self.invalid_args:
            arg, capitalized = capitalize(arg)
            if arg in val or capitalized in val:
                raise BadOption("Cannot specify arg in this statement", arg=arg, capitalized=capitalized, meta=meta)

    def make_spec(self):
        nsd = lambda spec: sb.defaulted(spec, NotSpecified)
        args = {}
        for arg, spec in self.args(self.self_type, self.self_name).items():
            arg, capitalized = capitalize(arg)
            args[(arg, capitalized)] = spec

        kwargs = {}
        for (arg, capitalized), spec in list(args.items()):
            kwargs[arg] = nsd(spec)
            kwargs[capitalized] = sb.any_spec()
        return args, sb.set_options(**kwargs)

    def make_kwargs(self, meta, args, normalised):
        kwargs = {}
        for (arg, capitalized) in args:
            if normalised.get(arg, NotSpecified) is not NotSpecified and normalised.get(capitalized, NotSpecified) is not NotSpecified:
                raise BadOption("Cannot specify arg as special and capitalized at the same time", arg=arg, special_val=normalised.get(arg), capitalized_val=normalised.get(capitalized), meta=meta)
            else:
                kwargs[arg] = normalised[capitalized] if normalised.get(capitalized, NotSpecified) is not NotSpecified else normalised[arg]
        return kwargs

    def complain_about_missing_args(self, meta, kwargs):
        missing = []
        for arg in self.required:
            if isinstance(arg, six.string_types):
                arg = (arg, )
            available = sorted(list(set(list(chain.from_iterable([capitalize(thing) for thing in arg])))))
            if not any(kwargs.get(option, NotSpecified) is not NotSpecified for option in available):
                missing.append(" or ".join(available))

        if missing:
            raise BadPolicy("Statement is missing required properties", missing=missing, meta=meta)

class resource_policy_dict(sb.Spec):
    def setup(self, effect=NotSpecified):
        self.effect = effect

    def normalise(self, meta, val):
        if isinstance(val, MergedOptions):
            val = val.as_dict()
        val = sb.dictionary_spec().normalise(meta, val)
        if self.effect is not NotSpecified:
            if val.get('effect', self.effect) != self.effect or val.get('Effect', self.effect) != self.effect:
                raise BadOption("Defaulted effect is being overridden", default=self.effect, overridden=val.get("Effect", val.get("effect")), meta=meta)

            if val.get('effect', NotSpecified) is NotSpecified and val.get("Effect", NotSpecified) is NotSpecified:
                val['Effect'] = self.effect
        return val

class permission_dict(resource_policy_dict):
    pass

class trust_dict(sb.Spec):
    def setup(self, principal):
        self.principal = principal

    def normalise(self, meta, val):
        if isinstance(val, MergedOptions):
            val = val.as_dict()
        val = sb.dictionary_spec().normalise(meta, val)
        if self.principal == "notprincipal":
            opposite = "principal"
        else:
            opposite = ("not", "principal")
        opposite, cap_opposite = capitalize(opposite)
        if opposite in val or cap_opposite in val:
            raise BadPolicy("Specifying opposite principal type in statement", wanted=self.principal, got=opposite, meta=meta)

        capitalized = self.principal
        if capitalized in ("principal", "Principal"):
            arg, capitalized = "principal", "Principal"
        else:
            arg, capitalized = "notprincipal", "NotPrincipal"

        if arg not in val and capitalized not in val:
            return {self.principal: val}
        return val

class permission_statement_spec(statement_spec):
    args = lambda s, self_type, self_name: {
          'sid': sb.string_spec()
        , 'effect': sb.string_choice_spec(choices=["Deny", "Allow"])

        , 'action': sb.listof(sb.string_spec())
        , ("not", "action"): sb.listof(sb.string_spec())

        , 'resource': resource_spec(self_type, self_name)
        , ('not', 'resource'): resource_spec(self_type, self_name)

        , 'condition': sb.dictionary_spec()
        , ('not', 'condition'): sb.dictionary_spec()
        }
    validators = [
          validators.deprecated_key('allow', "Use 'effect: Allow|Deny' instead")
        , validators.deprecated_key('disallow', "Use 'effect: Allow|Deny' instead")
        ]
    required = [('action', ('not', 'action')), 'effect', ('resource', ('not', 'resource'))]
    invalid_args = ['principal', ('not', 'principal')]
    final_kls = lambda s, *args, **kwargs: PermissionStatement(*args, **kwargs)

class resource_policy_statement_spec(statement_spec):
    args = lambda s, self_type, self_name: {
          'sid': sb.string_spec()

        , 'effect': sb.string_choice_spec(choices=["Deny", "Allow"])

        , 'action': sb.listof(sb.string_spec())
        , ("not", "action"): sb.listof(sb.string_spec())

        , 'resource': resource_spec(self_type, self_name)
        , ('not', 'resource'): resource_spec(self_type, self_name)

        , 'principal': sb.listof(principal_spec(self_type, self_name))
        , ('not', 'principal'): sb.listof(principal_spec(self_type, self_name))

        , 'condition': sb.dictionary_spec()
        , ('not', 'condition'): sb.dictionary_spec()
        }
    validators = [
          validators.deprecated_key('allow', "Use 'effect: Allow|Deny' instead")
        , validators.deprecated_key('disallow', "Use 'effect: Allow|Deny' instead")
        ]
    final_kls = lambda s, *args, **kwargs: ResourcePolicyStatement(*args, **kwargs)

class grant_statement_spec(statement_spec):
    args = lambda s, self_type, self_name: {
          'grantee': sb.required(resource_spec(self_type, self_name, only="iam"))
        , 'retiree': resource_spec(self_type, self_name, only="iam")
        , 'operations': sb.required(sb.listof(sb.string_spec()))
        , 'constraints': sb.any_spec()
        , 'grant_tokens': sb.any_spec()
        }
    final_kls = lambda s, *args, **kwargs: GrantStatement(*args, **kwargs)

class trust_statement_spec(resource_policy_statement_spec):
    final_kls = lambda s, *args, **kwargs: TrustStatement(*args, **kwargs)

class principal_service_spec(sb.Spec):
    def normalise(self, meta, val):
        if val == "ec2":
            return "ec2.amazonaws.com"
        raise BadOption("Unknown special principal service", specified=val, meta=meta)

class principal_spec(sb.Spec):
    def setup(self, self_type, self_name):
        self.self_type = self_type
        self.self_name = self_name

    def normalise(self, meta, val):
        iam_spec = iam_specs(val, self.self_type, self.self_name)

        result = sb.set_options(
              Service = sb.listof(sb.string_spec())
            , Federated = sb.listof(sb.string_spec())
            , AWS = sb.listof(sb.string_spec())
            ).normalise(meta, val)

        special = sb.set_options(
              service = sb.listof(principal_service_spec())
            , federated = resource_spec(self.self_type, self.self_name)
            , iam = iam_spec
            ).normalise(meta, val)

        for arg, lst in special.items():
            capitalized = arg.capitalize()
            if arg == 'iam':
                capitalized = "AWS"
            result[capitalized].extend(lst)

        for key, val in list(result.items()):
            if not val:
                del result[key]
                continue

            # Amazon gets rid of the lists if only one item
            # And this mucks around with the diffing....
            if len(val) is 1:
                result[key] = val[0]
            else:
                result[key] = sorted(val)

        return result

class PermissionStatement(dictobj):
    fields = ['sid', 'effect', 'action', 'notaction', 'resource', 'notresource', 'condition', 'notcondition']

    @property
    def statement(self):
        statement = {
              "Sid": self.sid, "Effect": self.effect, "Action": self.action, "NotAction": self.notaction
            , "Resource": self.resource, "NotResource": self.notresource
            , "Condition": self.condition, "NotCondition": self.notcondition
            }

        for key, val in list(statement.items()):
            if val is NotSpecified:
                del statement[key]

        for thing in ("Action", "NotAction", "Resource", "NotResource", "Condition", "NotCondition"):
            if thing in statement and isinstance(statement[thing], list):
                if len(statement[thing]) == 1:
                    statement[thing] = statement[thing][0]
                else:
                    statement[thing] = sorted(statement[thing])

        return statement

class ResourcePolicyStatement(dictobj):
    fields = ['sid', 'effect', 'action', 'notaction', 'resource', 'notresource', 'principal', 'notprincipal', 'condition', 'notcondition']

    def merge_principal(self, val, key):
        if len(val[key]) == 1 and isinstance(val[key], list):
            val[key] = val[key][0]
            return val[key]

        if not isinstance(val[key], list):
            val[key] = [val[key]]

        result = {}
        for item in val[key]:
            for service, lst in item.items():
                if not isinstance(lst, list):
                    lst = [lst]
                if service in result:
                    result[service].extend(lst)
                else:
                    result[service] = lst

        for service, lst in list(result.items()):
            if len(lst) == 1:
                result[service] = lst[0]

        return result

    @property
    def statement(self):
        statement = {
              "Sid": self.sid, "Effect": self.effect, "Action": self.action, "NotAction": self.notaction
            , "Resource": self.resource, "NotResource": self.notresource
            , "Principal": self.principal, "NotPrincipal": self.notprincipal
            , "Condition": self.condition, "NotCondition": self.notcondition
            }

        for key, val in list(statement.items()):
            if val is NotSpecified:
                del statement[key]

        if "Sid" not in statement:
            statement["Sid"] = ""

        if "Effect" not in statement:
            statement["Effect"] = "Allow"

        for principal in ("Principal", "NotPrincipal"):
            if principal in statement:
                statement[principal] = self.merge_principal(statement, principal)

        for thing in ("Action", "NotAction", "Resource", "NotResource", "Condition", "NotCondition", "Principal", "NotPrincipal"):
            if thing in statement and isinstance(statement[thing], list):
                if len(statement[thing]) == 1:
                    statement[thing] = statement[thing][0]
                else:
                    statement[thing] = sorted(statement[thing])

        return statement

class TrustStatement(ResourcePolicyStatement):

    @property
    def statement(self):
        statement = super(TrustStatement, self).statement

        if "Action" not in statement and 'NotAction' not in statement:
            if "Principal" in statement or "NotPrincipal" in statement:
                have_federated = "Principal" in statement and "Federated" in statement["Principal"] or "NotPrincipal" in statement and "Federated" in statement["NotPrincipal"]
                if have_federated:
                    statement["Action"] = "sts:AssumeRoleWithSAML"
                else:
                    statement["Action"] = "sts:AssumeRole"

        return statement

class GrantStatement(dictobj):
    fields = ['grantee', 'retiree', 'operations', 'grant_tokens', 'constraints']

    @property
    def statement(self):
        operations = self.operations
        if operations is not NotSpecified:
            operations = sorted(self.operations)
        statement = {
              "GranteePrincipal": self.grantee, "RetireePrincipal": self.retiree
            , "Operations": self.operations, "GrantTokens": self.grant_tokens
            , "Constraints": self.constraints
            }

        for key, val in list(statement.items()):
            if val is NotSpecified:
                del statement[key]

        for thing in ("GranteePrincipal", "RetireePrincipal"):
            if thing in statement and isinstance(statement[thing], list):
                if len(statement[thing]) == 1:
                    statement[thing] = statement[thing][0]
                else:
                    statement[thing] = sorted(statement[thing])

        return statement




"""
Here we define the yaml specification for aws_syncr options

The specifications are responsible for sanitation, validation and normalisation.
"""

from aws_syncr.option_spec.encryption_keys import encryption_keys_spec, EncryptionKeys
from aws_syncr.option_spec.buckets import buckets_spec, Buckets
from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.option_spec.roles import role_spec, Roles
from aws_syncr.errors import BadOption

from input_algorithms.spec_base import (
      defaulted, boolean, string_spec, formatted, create_spec, dictionary_spec
    , directory_spec, dictof, string_or_int_as_string_spec, container_spec
    )
from input_algorithms.validators import Validator
from input_algorithms.dictobj import dictobj

import six
import re

regexes = {
      "amazon_account_id": re.compile('\d{12}')
    }

class AwsSyncr(dictobj):
    fields = {
          "debug": "Set debug capability"
        , "dry_run": "Whether to do a dry run or not"
        , "extra": "Sets the ``$@`` variable. Alternatively specify these after a ``--`` on the commandline"
        , "stage": "Stage to deploy for an api gateway when deploy_gateway is used"
        , "location": "The location to base everything in"
        , "artifact": "Arbitrary argument"
        , "environment": "The environment to sync"
        , "config_folder": "The folder where the configuration can be found"
        }

class valid_account_id(Validator):
    def validate(self, meta, val):
        """Validate an account_id"""
        val = string_or_int_as_string_spec().normalise(meta, val)
        if not regexes['amazon_account_id'].match(val):
            raise BadOption("Account id must match a particular regex", got=val, should_match=regexes['amazon_account_id'].pattern)
        return val

class AwsSyncrSpec(object):
    """Knows about aws_syncr specific configuration"""

    @property
    def aws_syncr_spec(self):
        """Spec for aws_syncr options"""
        formatted_string = formatted(string_spec(), MergedOptionStringFormatter, expected_type=six.string_types)
        return create_spec(AwsSyncr
            , extra = defaulted(formatted_string, "")
            , stage = defaulted(formatted_string, "")
            , debug = defaulted(boolean(), False)
            , dry_run = defaulted(boolean(), False)
            , location = defaulted(formatted_string, "ap-southeast-2")
            , artifact = formatted_string
            , environment = formatted_string
            , config_folder = directory_spec()
            )

    @property
    def accounts_spec(self):
        """Spec for accounts options"""
        formatted_account_id = formatted(valid_account_id(), MergedOptionStringFormatter, expected_type=six.string_types)
        return dictof(string_spec(), formatted_account_id)

    @property
    def templates_spec(self):
        """Spec for templates"""
        return dictof(string_spec(), dictionary_spec())




from aws_syncr.filename_completer import filename_prompt, setup_completer
from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.errors import AwsSyncrError
from aws_syncr.errors import UserQuit

from Crypto.Util import Counter
from Crypto.Cipher import AES

from option_merge import MergedOptions
from six.moves import input
from textwrap import dedent
import itertools
import readline
import logging
import base64
import shlex
import yaml
import sys
import six
import os

log = logging.getLogger("aws_syncr.actions")

available_actions = {}

def an_action(func):
    available_actions[func.__name__] = func
    func.label = "Default"
    return func

def find_lambda_function(aws_syncr, configuration):
    lambda_function = aws_syncr.artifact

    if 'lambda' not in configuration:
        raise AwsSyncrError("Please define lambda functions under the 'lambda' section of your configuration")

    if not lambda_function:
        raise AwsSyncrError("Please specify --artifact for the lambda function to deploy")

    wanted = ['lambda', lambda_function]
    if wanted not in configuration:
        raise AwsSyncrError("Couldn't find specified lambda function", available=list(configuration["lambda"].items.keys()))

    return configuration['lambda'].items[lambda_function]

def find_gateway(aws_syncr, configuration):
    amazon = configuration['amazon']

    stage = aws_syncr.stage
    gateway = aws_syncr.artifact

    if 'apigateway' not in configuration:
        raise AwsSyncrError("Please define apigateway in your configuration before trying to deploy a gateway")

    if not gateway:
        available = list(configuration['apigateway'].items.keys())
        raise AwsSyncrError("Please specify --artifact for the gateway function to deploy", available=available)

    wanted = ['apigateway', gateway]
    if wanted not in configuration:
        raise AwsSyncrError("Couldn't find specified api gateway", available=list(configuration["apigateway"].items.keys()))
    gateway = configuration['apigateway'].items[gateway]

    if not stage:
        raise AwsSyncrError("Please specify --stage", available=list(gateway.stage_names))

    return aws_syncr, amazon, stage, gateway

def find_certificate_source(configuration, gateway, certificate):
    source = configuration.source_for(['apigateway', gateway, 'domain_names'])
    location = ["apigateway", gateway, 'domain_names']
    domain_names = configuration.get(location, ignore_converters=True)

    for name, domain in domain_names.items():
        if 'zone' in domain:
            zone = MergedOptionStringFormatter(configuration, '.'.join(location + ['zone']), value=domain.get('zone')).format()
            domain_name = "{0}.{1}".format(name, zone)
            if domain_name == certificate:
                if 'certificate' not in domain:
                    domain['certificate'] = {}

                var = domain['certificate']

                class StickyChain(object):
                    def __init__(self):
                        self.lst = []

                    def __add__(self, other):
                        self.lst.extend(other)
                        return self.lst

                    def __contains__(self, item):
                        return item in self.lst

                    def __getitem__(self, index):
                        return self.lst[index]
                chain = StickyChain()

                if isinstance(var, six.string_types):
                    result = MergedOptionStringFormatter(configuration, '.'.join(location), value=var, chain=chain).format()
                    if not isinstance(result, dict) and not isinstance(result, MergedOptions) and (not hasattr(result, 'is_dict') or not result.is_dict):
                        raise AwsSyncrError("certificate should be pointing at a dictionary", got=result, chain=['.'.join(location)] + chain)

                    location = chain[-1]
                    source = configuration.source_for(location)
                    for info in configuration.storage.get_info(location, ignore_converters=True):
                        location = [str(part) for part in info.path.path]

    return location, source

@an_action
def list_tasks(collector):
    """List the available_tasks"""
    print("Usage: aws_syncr <environment_folder> <task>")
    print("")
    print("Available environment folders to choose from are")
    print("------------------------------------------------")
    print("")
    for environment in os.listdir(collector.configuration_folder):
        location = os.path.join(collector.configuration_folder, environment)
        if os.path.isdir(location) and not environment.startswith("."):
            print("\t{0}".format(location))

    print("")
    print("Available tasks to choose from are:")
    print("-----------------------------------")
    print("")
    keygetter = lambda item: item[1].label
    tasks = sorted(available_actions.items(), key=keygetter)
    sorted_tasks = sorted(list(tasks), key=lambda item: len(item[0]))
    max_length = max(len(name) for name, _ in sorted_tasks)
    for key, task in sorted_tasks:
        desc = dedent(task.__doc__ or "").strip().split('\n')[0]
        print("\t{0}{1} :-: {2}".format(" " * (max_length-len(key)), key, desc))
    print("")

@an_action
def sync(collector):
    """Sync an environment"""
    amazon = collector.configuration['amazon']
    aws_syncr = collector.configuration['aws_syncr']

    # Convert everything before we try and sync anything
    log.info("Converting configuration")
    converted = {}
    for thing in collector.configuration["__registered__"]:
        if thing in collector.configuration:
            converted[thing] = collector.configuration[thing]

    # Do the sync
    for typ in collector.configuration["__registered__"]:
        if typ in converted:
            thing = converted[typ]
            if not aws_syncr.artifact or aws_syncr.artifact == typ:
                log.info("Syncing {0}".format(typ))
                for name, item in thing.items.items():
                    thing.sync_one(aws_syncr, amazon, item)

    if not amazon.changes:
        log.info("No changes were made!!")

@an_action
def deploy_lambda(collector):
    """Deploy a lambda function"""
    amazon = collector.configuration['amazon']
    aws_syncr = collector.configuration['aws_syncr']
    find_lambda_function(aws_syncr, collector.configuration).deploy(aws_syncr, amazon)

@an_action
def test_lambda(collector):
    amazon = collector.configuration['amazon']
    amazon._validated = True
    aws_syncr = collector.configuration['aws_syncr']
    if not find_lambda_function(aws_syncr, collector.configuration).test(aws_syncr, amazon):
        raise AwsSyncrError("Failed to test the lambda")

@an_action
def deploy_and_test_lambda(collector):
    deploy_lambda(collector)
    test_lambda(collector)

@an_action
def deploy_gateway(collector):
    configuration = collector.configuration
    aws_syncr = configuration['aws_syncr']
    aws_syncr, amazon, stage, gateway = find_gateway(aws_syncr, configuration)
    gateway.deploy(aws_syncr, amazon, stage)

    if not configuration['amazon'].changes:
        log.info("No changes were made!!")

@an_action
def sync_and_deploy_gateway(collector):
    configuration = collector.configuration
    aws_syncr = configuration['aws_syncr']
    find_gateway(aws_syncr, configuration)

    artifact = aws_syncr.artifact
    aws_syncr.artifact = ""
    sync(collector)

    aws_syncr.artifact = artifact
    deploy_gateway(collector)

@an_action
def test_gateway(collector):
    collector.configuration['amazon']._validated = True
    configuration = collector.configuration
    aws_syncr = configuration['aws_syncr']
    aws_syncr, amazon, stage, gateway = find_gateway(aws_syncr, configuration)
    if not gateway.test(aws_syncr, amazon, stage):
        raise AwsSyncrError("Failed to test the gateway")

@an_action
def test_all_gateway_endpoints(collector):
    collector.configuration['amazon']._validated = True
    configuration = collector.configuration
    aws_syncr = configuration['aws_syncr']
    aws_syncr, amazon, stage, gateway = find_gateway(aws_syncr, configuration)

    failure = False
    for method, resource in gateway.available_methods_and_endpoints():
        combination = "{0} {1}".format(method, resource)
        print(combination)
        print("=" * len(combination))
        aws_syncr.extra = combination
        if not gateway.test(aws_syncr, amazon, stage):
            failure = True
        print("")

    if failure:
        raise AwsSyncrError("Atleast one of the endpoints failed the test")

@an_action
def encrypt_certificate(collector):
    configuration = collector.configuration
    amazon = configuration['amazon']
    aws_syncr = configuration['aws_syncr']
    certificate = aws_syncr.artifact

    available = []

    for gateway_name, gateway in configuration.get('apigateway', {}, ignore_converters=True).items():
        for name, options in gateway.get("domain_names", {}).items():
            if "zone" in options:
                location = '.'.join(['apigateway', gateway_name, 'domain_names'])
                formatter = MergedOptionStringFormatter(configuration, location, value=options['zone'])
                available.append((gateway_name, "{0}.{1}".format(name, formatter.format())))

    if not available:
        raise AwsSyncrError("Please specify apigateway.<gateway_name>.domain_names.<domain_name>.name in the configuration")

    if not certificate:
        raise AwsSyncrError("Please specify certificate to encrypt with --artifact", available=[a[1] for a in available])

    if certificate not in [a[1] for a in available]:
        raise AwsSyncrError("Unknown certificate", available=[a[1] for a in available], got=certificate)

    gateway = [name for name, cert in available if cert == certificate][0]
    location, source = find_certificate_source(configuration, gateway, certificate)

    log.info("Gonna edit {0} in {1}".format(location, source))
    current = MergedOptions.using(yaml.load(open(source)))
    dest = current[location]

    try:
        key_id = input("Which kms key do you want to use? ")
        region = input("What region is this key in? ")
    except EOFError:
        raise UserQuit()

    # Make the filename completion work
    setup_completer()

    # Create the datakey to encrypt with
    data_key = amazon.kms.generate_data_key(region, key_id)
    plaintext_data_key = data_key["Plaintext"]
    encrypted_data_key = base64.b64encode(data_key["CiphertextBlob"]).decode('utf-8')

    # Encrypt our secrets
    secrets = {}
    for name, desc in (("body", "certificate's crt file"), ("key", "private key file"), ("chain", "certificate chain")):
        location = None
        while not location or not os.path.isfile(location):
            location = os.path.expanduser(filename_prompt("Where is the {0}? ".format(desc)))
            if not location or not os.path.isfile(location):
                print("Please give a location to a file that exists!")

        data = open(location).read()
        counter = Counter.new(128)
        encryptor = AES.new(plaintext_data_key[:32], AES.MODE_CTR, counter=counter)
        secrets[name] = base64.b64encode(encryptor.encrypt(data)).decode('utf-8')

    # Add in the encrypted values
    dest['body'] = {"kms": secrets['body'], "location": region, "kms_data_key": encrypted_data_key}
    dest['key'] = {"kms": secrets['key'], "location": region, "kms_data_key": encrypted_data_key}
    dest['chain'] = {"kms": secrets['chain'], "location": region, "kms_data_key": encrypted_data_key}

    # And write to the file!
    yaml.dump(current.as_dict(), open(source, 'w'), explicit_start=True, indent=2, default_flow_style=False)

@an_action
def execute_as(collector):
    """Execute a command (after the --) as an assumed role (specified by --artifact)"""
    # Gonna assume role anyway...
    collector.configuration['amazon']._validated = True

    # Find the arn we want to assume
    account_id = collector.configuration['accounts'][collector.configuration['aws_syncr'].environment]
    arn = "arn:aws:iam::{0}:role/{1}".format(account_id, collector.configuration['aws_syncr'].artifact)

    # Determine the command to run
    parts = shlex.split(collector.configuration["aws_syncr"].extra)
    if not parts:
        suggestion = " ".join(sys.argv) + " -- /path/to/command_to_run"
        msg = "No command was provided. Try something like:\n\t\t{0}".format(suggestion)
        raise AwsSyncrError(msg)

    # Get our aws credentials environment variables from the assumed role
    env = dict(os.environ)
    env.update(collector.configuration['amazon'].iam.assume_role_credentials(arn))

    # Turn into the command we want to execute
    os.execvpe(parts[0], parts, env)




"""
Collects then parses configuration files and verifies that they are valid.
"""

from aws_syncr.errors import BadConfiguration, BadYaml, BadOption, BadImport
from aws_syncr.option_spec.aws_syncr_specs import AwsSyncrSpec
from aws_syncr.amazon import Amazon

from input_algorithms.dictobj import dictobj
from input_algorithms.meta import Meta

from option_merge.collector import Collector
from option_merge import MergedOptions
from option_merge import Converter

import pkg_resources
import tempfile
import logging
import yaml
import json
import imp
import os

log = logging.getLogger("aws_syncr.collector")

class Collector(Collector):

    BadFileErrorKls = BadYaml
    BadConfigurationErrorKls = BadConfiguration

    def alter_clone_cli_args(self, new_collector, new_cli_args, new_aws_syncr_options=None):
        new_aws_syncr = self.configuration["aws_syncr"].clone()
        if new_aws_syncr_options:
            new_aws_syncr.update(new_aws_syncr_options)
        new_cli_args["aws_syncr"] = new_aws_syncr

    def prepare(self, configuration_folder, cli_args, environment):
        """Make a temporary configuration file from the files in our folder"""
        self.configuration_folder = configuration_folder
        if not os.path.isdir(configuration_folder):
            raise BadOption("Specified configuration folder is not a directory!", wanted=configuration_folder)
        available = [os.path.join(configuration_folder, name) for name in os.listdir(configuration_folder)]
        available_environments = [os.path.abspath(path) for path in available if os.path.isdir(path)]

        if environment and os.path.abspath(environment) not in available_environments:
            raise BadOption("Specified environment doesn't exist", available=available_environments, wanted=environment)

        if environment:
            environment_files = [os.path.join(environment, "../accounts.yaml")]
            for root, dirs, files in os.walk(environment):
                environment_files.extend(os.path.join(root, filename) for filename in files)

            with tempfile.NamedTemporaryFile() as fle:
                contents = json.dumps({"includes": environment_files})
                fle.write(contents.encode('utf-8'))
                fle.flush()
                cli_args['aws_syncr']['environment'] = os.path.split(environment)[-1]
                super(Collector, self).prepare(fle.name, cli_args)

    def find_missing_config(self, configuration):
        """Complain if we have no account information"""
        if "accounts" not in configuration:
            raise self.BadConfigurationErrorKls("accounts is a mandatory section and wasn't found")

    def extra_prepare(self, configuration, cli_args):
        """Called before the configuration.converters are activated"""
        aws_syncr = cli_args.pop("aws_syncr")

        self.configuration.update(
            { "$@": aws_syncr.get("extra", "")
            , "aws_syncr": aws_syncr
            , "templates": {}
            , "config_folder": self.configuration_folder
            }
        , source = "<cli_args>"
        )

    def extra_prepare_after_activation(self, configuration, cli_args):
        """Setup our connection to amazon"""
        aws_syncr = configuration['aws_syncr']
        configuration["amazon"] = Amazon(configuration['aws_syncr'].environment, configuration['accounts'], debug=aws_syncr.debug, dry_run=aws_syncr.dry_run)

    def home_dir_configuration_location(self):
        return os.path.expanduser("~/.aws_syncrrc.yml")

    def start_configuration(self):
        """Create the base of the configuration"""
        return MergedOptions(dont_prefix=[dictobj])

    def read_file(self, location):
        """Read in a yaml file and return as a python object"""
        try:
            return yaml.load(open(location))
        except (yaml.parser.ParserError, yaml.scanner.ScannerError) as error:
            raise self.BadFileErrorKls("Failed to read yaml", location=location, error_type=error.__class__.__name__, error="{0}{1}".format(error.problem, error.problem_mark))

    def add_configuration(self, configuration, collect_another_source, done, result, src):
        """Used to add a file to the configuration, result here is the yaml.load of the src"""
        if "includes" in result:
            for include in result["includes"]:
                collect_another_source(include)
        configuration.update(result, source=src)

    def extra_configuration_collection(self, configuration):
        """Hook to do any extra configuration collection or converter registration"""
        aws_syncr_spec = AwsSyncrSpec()
        registered = {}
        directory = pkg_resources.resource_filename("aws_syncr", "option_spec")

        for location in sorted(os.listdir(directory)):
            import_name = os.path.splitext(location)[0]
            if import_name != '__pycache__':
                try:
                    args = imp.find_module(import_name, [directory])
                except ImportError as error:
                    raise BadImport(directory=directory, importing=import_name, error=error)

                try:
                    module = imp.load_module(import_name, *args)
                except SyntaxError as error:
                    raise BadImport(directory=self.directory, importing=self.import_name, error=error)

                if hasattr(module, "__register__"):
                    registered.update(module.__register__())

        configuration['__registered__'] = [name for _, name in sorted(registered.keys())]
        by_name = dict((r[1], registered[r]) for r in registered)
        for thing in ['aws_syncr', 'accounts', 'templates'] + list(by_name.keys()):
            def make_converter(thing):
                def converter(p, v):
                    log.info("Converting %s", p)
                    meta = Meta(p.configuration, [(thing, "")])
                    configuration.converters.started(p)
                    if thing in by_name:
                        return by_name[thing].normalise(meta, v)
                    else:
                        return getattr(aws_syncr_spec, "{0}_spec".format(thing)).normalise(meta, v)
                return converter
            configuration.add_converter(Converter(convert=make_converter(thing), convert_path=[thing]))




from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.errors import BadTemplate

from input_algorithms.errors import BadSpecValue
from input_algorithms.dictobj import dictobj
from input_algorithms import spec_base as sb

from option_merge import MergedOptions
import six

class route_spec(sb.Spec):
    def normalise(self, meta, val):
        if 'use' in val:
            template = val['use']
            if template not in meta.everything['templates']:
                available = list(meta.everything['templates'].keys())
                raise BadTemplate("Template doesn't exist!", wanted=template, available=available, meta=meta)

            val = MergedOptions.using(meta.everything['templates'][template], val)

        formatted_string = sb.formatted(sb.string_spec(), MergedOptionStringFormatter)
        route_name = meta.key_names()['_key_name_0']

        val = sb.create_spec(DNSRoute
            , name = sb.overridden(route_name)
            , zone = formatted_string
            , record_type = sb.string_choice_spec(["CNAME"])
            , record_target = formatted_string
            ).normalise(meta, val)

        if not val.zone.endswith("."):
            val.zone = "{0}.".format(val.zone)

        if not isinstance(val.record_target, six.string_types):
            if not hasattr(val.record_target, "cname"):
                raise BadSpecValue("record_target must point at an object with a cname property", got=type(val.record_target), meta=meta)
            val.record_target = val.record_target.cname

        return val

class DNSRoutes(dictobj):
    fields = ['items']

    def sync_one(self, aws_syncr, amazon, route):
        """Make sure this role exists and has only what policies we want it to have"""
        route_info = amazon.route53.route_info(route.name, route.zone)
        target = route.record_target
        if callable(target):
            target = target(amazon)

        if not route_info:
            amazon.route53.create_route(route.name, route.zone, route.record_type, target)
        else:
            amazon.route53.modify_route(route_info, route.name, route.zone, route.record_type, target)

class DNSRoute(dictobj):
    fields = {
        "name": "The name of the record"
      , "zone": "The zone this record sits in"
      , "record_type": "The type of the record"
      , "record_target": "Where the record points at"
      }

def __register__():
    return {(100, "dns"): sb.container_spec(DNSRoutes, sb.dictof(sb.string_spec(), route_spec()))}




from aws_syncr.errors import UserQuit

from six.moves import input
import readline
import glob
import os

def setup_completer():
    if 'libedit' in readline.__doc__:
        readline.parse_and_bind("bind ^I rl_complete")
    else:
        readline.parse_and_bind("tab: complete")
    readline.set_completer(NoneCompleter)

class NoneCompleter(object):
    def complete(self, text, state):
        self.matches = []
        return None

class FilenameCompleter(object):
    def complete(self, text, state):
        text = os.path.expanduser(readline.get_line_buffer())
        if not text:
            self.matches = glob.glob("*")
        else:
            if os.path.isfile(text):
                self.matches = []

            else:
                dirname = os.path.dirname(text)
                if not dirname:
                    self.matches = glob.glob("*")
                else:
                    self.matches = glob.glob(os.path.join(dirname, "*"))

            self.matches = [os.path.basename(match) for match in self.matches if match.startswith(text)]

        if len(self.matches) > state:
            if len(self.matches) == 1:
                if os.path.isdir(os.path.join(os.path.dirname(text), self.matches[state])):
                    return "{0}/".format(self.matches[state])
            return self.matches[state]
        else:
            return None

def custom_prompt(msg, delims="", completer=lambda: None):
    """Start up a prompt that with particular delims and completer"""
    try:
        orig_delims = readline.get_completer_delims()
        orig_completer = readline.get_completer()

        readline.set_completer_delims(delims)
        readline.set_completer(completer)

        try:
            ret = input(msg)
        finally:
            readline.set_completer_delims(orig_delims)
            readline.set_completer(orig_completer)

        return ret
    except EOFError:
        raise UserQuit()

def filename_prompt(msg, delims="/"):
    completer = FilenameCompleter().complete
    return custom_prompt(msg, delims, completer)




from setuptools import setup, find_packages
from aws_syncr import VERSION

setup(
      name = "aws_syncr"
    , version = VERSION
    , packages = ['aws_syncr'] + ['aws_syncr.%s' % pkg for pkg in find_packages('aws_syncr')]
    , include_package_data = True

    , install_requires =
      [ "delfick_app==0.6.8"
      , "option_merge==0.9.9.2.2"
      , "input_algorithms==0.4.5"

      , "six"
      , "datadiff"
      , "requests"

      , "boto3==1.2.1"
      , "pyYaml==3.10"
      , 'pycrypto==2.6.1'
      ]

    , extras_require =
      { "tests":
        [ "noseOfYeti>=1.5.0"
        , "nose"
        , "mock==1.0.1"
        , "tox"
        ]
      }

    , entry_points =
      { 'console_scripts' :
        [ 'aws_syncr = aws_syncr.executor:main'
        ]
      }

    # metadata for upload to PyPI
    , url = "https://github.com/delfick/aws_syncr"
    , author = "Stephen Moore"
    , author_email = "stephen.moore@rea-group.com"
    , description = "Application that reads yaml and syncs definitions with amazon"
    , license = "WTFPL"
    , keywords = "aws"
    )




from aws_syncr.option_spec.statements import trust_statement_spec, permission_statement_spec, permission_dict, trust_dict
from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.option_spec.documents import Document
from aws_syncr.errors import BadOption, BadTemplate

from input_algorithms.spec_base import NotSpecified
from input_algorithms.dictobj import dictobj
from input_algorithms import spec_base as sb

from option_merge import MergedOptions
import logging
import six

log = logging.getLogger("aws_syncr.option_spec.roles")

class role_spec(object):
    def normalise(self, meta, val):
        if 'use' in val:
            template = val['use']
            if template not in meta.everything['templates']:
                available = list(meta.everything['templates'].keys())
                raise BadTemplate("Template doesn't exist!", wanted=template, available=available, meta=meta)

            val = MergedOptions.using(meta.everything['templates'][template], val)

        formatted_string = sb.formatted(sb.string_spec(), MergedOptionStringFormatter, expected_type=six.string_types)
        role_name = meta.key_names()['_key_name_0']

        original_permission = sb.listof(permission_dict()).normalise(meta.at("permission"), NotSpecified if "permission" not in val else val["permission"])
        deny_permission = sb.listof(permission_dict(effect='Deny')).normalise(meta.at("deny_permission"), NotSpecified if "deny_permission" not in val else val["deny_permission"])
        allow_permission = sb.listof(permission_dict(effect='Allow')).normalise(meta.at("allow_permission"), NotSpecified if "allow_permission" not in val else val["allow_permission"])

        allow_to_assume_me = sb.listof(trust_dict("principal")).normalise(meta.at("allow_to_assume_me"), val.get("allow_to_assume_me", NotSpecified))
        disallow_to_assume_me = sb.listof(trust_dict("notprincipal")).normalise(meta.at("disallow_to_assume_me"), val.get("disallow_to_assume_me", NotSpecified))

        val = val.wrapped()
        val['trust'] = allow_to_assume_me + disallow_to_assume_me
        val['permission'] = original_permission + deny_permission + allow_permission
        return sb.create_spec(Role
            , name = sb.overridden(role_name)
            , description = formatted_string
            , trust = sb.container_spec(Document, sb.listof(trust_statement_spec('role', role_name)))
            , permission = sb.container_spec(Document, sb.listof(permission_statement_spec('role', role_name)))
            , make_instance_profile = sb.defaulted(sb.boolean(), False)
            ).normalise(meta, val)

class Roles(dictobj):
    fields = ['items']

    def sync_one(self, aws_syncr, amazon, role):
        """Make sure this role exists and has only what policies we want it to have"""
        trust_document = role.trust.document
        permission_document = role.permission.document
        policy_name = "syncr_policy_{0}".format(role.name.replace('/', '__'))

        role_info = amazon.iam.role_info(role.name)
        if not role_info:
            amazon.iam.create_role(role.name, trust_document, policies={policy_name: permission_document})
        else:
            amazon.iam.modify_role(role_info, role.name, trust_document, policies={policy_name: permission_document})

        if role.make_instance_profile:
            amazon.iam.make_instance_profile(role.name)

class Role(dictobj):
    fields = {
        "name": "The name of the role"
      , "description": "The description of the role!"
      , "make_instance_profile": "Whether to make an instance profile for this role as well"

      , "trust": "The trust document"
      , "permission": "Combination of allow_permission and deny_permission"
      }

def __register__():
    return {(21, "roles"): sb.container_spec(Roles, sb.dictof(sb.string_spec(), role_spec()))}




from aws_syncr.option_spec.statements import resource_policy_statement_spec, resource_policy_dict
from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.option_spec.documents import Document
from aws_syncr.errors import BadTemplate

from input_algorithms.spec_base import NotSpecified
from input_algorithms import spec_base as sb
from input_algorithms.spec_base import Spec
from input_algorithms.dictobj import dictobj

from option_merge import MergedOptions
import six

class buckets_spec(Spec):
    def normalise(self, meta, val):
        if 'use' in val:
            template = val['use']
            if template not in meta.everything['templates']:
                available = list(meta.everything['templates'].keys())
                raise BadTemplate("Template doesn't exist!", wanted=template, available=available, meta=meta)

            val = MergedOptions.using(meta.everything['templates'][template], val)

        formatted_string = sb.formatted(sb.string_or_int_as_string_spec(), MergedOptionStringFormatter, expected_type=six.string_types)
        bucket_name = meta.key_names()['_key_name_0']

        original_permission = sb.listof(resource_policy_dict()).normalise(meta.at("permission"), NotSpecified if "permission" not in val else val["permission"])
        deny_permission = sb.listof(resource_policy_dict(effect='Deny')).normalise(meta.at("deny_permission"), NotSpecified if "deny_permission" not in val else val["deny_permission"])
        allow_permission = sb.listof(resource_policy_dict(effect='Allow')).normalise(meta.at("allow_permission"), NotSpecified if "allow_permission" not in val else val["allow_permission"])

        val = val.wrapped()
        val['permission'] = original_permission + deny_permission + allow_permission
        return sb.create_spec(Bucket
            , name = sb.overridden(bucket_name)
            , location = sb.required(formatted_string)
            , permission = sb.container_spec(Document, sb.listof(resource_policy_statement_spec('bucket', bucket_name)))
            , tags = sb.dictof(sb.string_spec(), formatted_string)
            ).normalise(meta, val)

class Buckets(dictobj):
    fields = ['items']

    def sync_one(self, aws_syncr, amazon, bucket):
        """Make sure this bucket exists and has only attributes we want it to have"""
        if bucket.permission.statements:
            permission_document = bucket.permission.document
        else:
            permission_document = ""

        bucket_info = amazon.s3.bucket_info(bucket.name)
        if not bucket_info:
            amazon.s3.create_bucket(bucket.name, permission_document, bucket.location, bucket.tags)
        else:
            amazon.s3.modify_bucket(bucket_info, bucket.name, permission_document, bucket.location, bucket.tags)

class Bucket(dictobj):
    fields = {
          'name': "Name of the bucket"
        , 'location': "The region the bucket exists in"
        , 'permission': "The permission statements to attach to the bucket"
        , 'tags': "The tags to associate with the bucket"
        }

def __register__():
    return {(80, "buckets"): sb.container_spec(Buckets, sb.dictof(sb.string_spec(), buckets_spec()))}




from aws_syncr.amazon.common import AmazonMixin
from aws_syncr.differ import Differ

from botocore.exceptions import ClientError
import boto3

import logging
import json
import os

log = logging.getLogger("aws_syncr.amazon.iam")

class Iam(AmazonMixin, object):
    def __init__(self, amazon, environment, accounts, dry_run):
        self.amazon = amazon
        self.dry_run = dry_run

        self.accounts = accounts
        self.account_id = accounts[environment]
        self.environment = environment

        self.client = self.amazon.session.client("iam")
        self.resource = self.amazon.session.resource('iam')

    def role_info(self, role_name):
        role = self.resource.Role(role_name.split('/')[-1])
        with self.ignore_missing():
            role.load()
            return role

    def create_role(self, name, trust_document, policies):
        with self.catch_boto_400("Couldn't Make role", "{0} assume document".format(name), trust_document, role=name):
            for _ in self.change("+", "role", role=name, document=trust_document):
                self.resource.create_role(Path="/{0}/".format('/'.join(name.split('/')[:-1])), RoleName=name.split('/')[-1], AssumeRolePolicyDocument=trust_document)

        if policies:
            for policy_name, document in policies.items():
                if document:
                    with self.catch_boto_400("Couldn't add policy", "{0} - {1} Permission document".format(name, policy_name), document, role=name, policy_name=policy_name):
                        for _ in self.change("+", "role_policy", role=name, policy=policy_name, document=document):
                            self.resource.RolePolicy(name.split('/')[-1], policy_name).put(PolicyDocument=document)

    def modify_role(self, role_info, name, trust_document, policies):
        changes = list(Differ.compare_two_documents(json.dumps(role_info.assume_role_policy_document), trust_document))
        if changes:
            with self.catch_boto_400("Couldn't modify trust document", "{0} assume document".format(name), trust_document, role=name):
                for _ in self.change("M", "trust_document", role=name, changes=changes):
                    self.resource.AssumeRolePolicy(name.split('/')[-1]).update(PolicyDocument=trust_document)

        with self.catch_boto_400("Couldn't get policies for a role", role=name):
            current_policies = dict((policy.name, policy) for policy in role_info.policies.all())
        unknown = [key for key in current_policies if key not in policies]

        if unknown:
            log.info("Role has unknown policies that will be disassociated\trole=%s\tunknown=%s", name, unknown)
            for policy in unknown:
                with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=name):
                    for _ in self.change("-", "role_policy", role=name, policy=policy):
                        current_policies[policy].delete()

        for policy, document in policies.items():
            if not document:
                if policy in current_policies:
                    with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=name):
                        for _ in self.change("-", "policy", role=name, policy=policy):
                            current_policies[policy].delete()
            else:
                needed = False
                changes = None

                if policy in current_policies:
                    changes = list(Differ.compare_two_documents(json.dumps(current_policies.get(policy).policy_document), document))
                    if changes:
                        log.info("Overriding existing policy\trole=%s\tpolicy=%s", name, policy)
                        needed = True
                else:
                    log.info("Adding policy to existing role\trole=%s\tpolicy=%s", name, policy)
                    needed = True

                if needed:
                    with self.catch_boto_400("Couldn't add policy document", "{0} - {1} policy document".format(name, policy), document, role=name, policy=policy):
                        symbol = "M" if changes else "+"
                        for _ in self.change(symbol, "role_policy", role=name, policy=policy, changes=changes, document=document):
                            if policy in current_policies:
                                current_policies[policy].put(PolicyDocument=document)
                            else:
                                self.client.put_role_policy(RoleName=name.split("/")[-1], PolicyName=policy, PolicyDocument=document)


    def make_instance_profile(self, name):
        role_name = name.split('/')[-1]
        existing_roles_in_profile = None
        with self.ignore_missing():
            existing_roles_in_profile = self.resource.InstanceProfile(role_name).roles

        if existing_roles_in_profile is None:
            try:
                with self.catch_boto_400("Couldn't create instance profile", instance_profile=name):
                    for _ in self.change("+", "instance_profile", profile=name):
                        self.resource.InstanceProfile(role_name).add_role(RoleName=role_name)
            except ClientError as error:
                if error.response["ResponseMetadata"]["HTTPStatusCode"] == 409:
                    # I'd rather ignore this conflict, than list all the instance_profiles
                    # Basically, the instance exists but isn't associated with the role
                    pass
                else:
                    raise

        if existing_roles_in_profile and any(rl.name != role_name for rl in existing_roles_in_profile):
            for role in [rl for rl in existing_roles_in_profile if rl.name != role_name]:
                with self.catch_boto_400("Couldn't remove role from an instance profile", profile=role_name, role=role):
                    for _ in self.change("-", "instance_profile_role", profile=role_name, role=role):
                        self.resource.InstanceProfile(role_name).remove_role(RoleName=role)

        if not existing_roles_in_profile or not any(rl.name == role_name for rl in existing_roles_in_profile):
            with self.catch_boto_400("Couldn't add role to an instance profile", role=name, instance_profile=role_name):
                for _ in self.change("+", "instance_profile_role", profile=role_name, role=role_name):
                    self.resource.InstanceProfile(role_name).add_role(RoleName=role_name)

    def assume_role_credentials(self, arn):
        """Return the environment variables for an assumed role"""
        log.info("Assuming role as %s", arn)

        # Clear out empty values
        for name in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_SECURITY_TOKEN', 'AWS_SESSION_TOKEN']:
            if name in os.environ and not os.environ[name]:
                del os.environ[name]

        sts = self.amazon.session.client("sts")
        with self.catch_boto_400("Couldn't assume role", arn=arn):
            creds = sts.assume_role(RoleArn=arn, RoleSessionName="aws_syncr")

        return {
              'AWS_ACCESS_KEY_ID': creds["Credentials"]["AccessKeyId"]
            , 'AWS_SECRET_ACCESS_KEY': creds["Credentials"]["SecretAccessKey"]
            , 'AWS_SECURITY_TOKEN': creds["Credentials"]["SessionToken"]
            , 'AWS_SESSION_TOKEN': creds["Credentials"]["SessionToken"]
            }

